import os

import jsonschema
import mock

from doctor.errors import (
    ForbiddenError, ImmutableError, InvalidValueError, ParseError,
    NotFoundError, SchemaValidationError, UnauthorizedError)
from doctor.flask import (
    FlaskResourceSchema, SchematicHTTPException, handle_http,
    HTTP400Exception, HTTP401Exception, HTTP403Exception, HTTP404Exception,
    HTTP409Exception, HTTP500Exception)
from doctor.response import Response
from .base import TestCase


class TestFlask(TestCase):

    def setUp(self):
        self.mock_handler = mock.Mock()

        self.mock_logic = mock.Mock(spec=['_autospec'])
        self.mock_logic.return_value = {'foo': 1}
        self.response_headers = {
            'Content-Type': 'text/csv',
            'Content-Disposition': 'attachement; filename=data.csv',
            'X-Custom-Response-Header': '12345',
        }
        self.mock_flask_response_logic = mock.Mock(spec=['_autospec'])
        self.mock_flask_response_logic.return_value = Response(
            '1,2\n3,4', self.response_headers)

        self.mock_request_patch = mock.patch('doctor.flask.request')
        self.mock_request = self.mock_request_patch.start()

        self.schema = FlaskResourceSchema({
            'definitions': {
                'a': {'type': 'string'},
                'b': {'type': 'integer'},
                'c': {'type': 'integer'},
                'response': {
                    'type': 'object',
                    'properties': {
                        'foo': {'type': 'integer'},
                    },
                    'additionalProperties': False
                }
            }
        })
        self.request_schema = self.schema._create_request_schema(
            params=('a', 'b', 'c'), required=('a', 'b'))
        self.request_validator = jsonschema.Draft4Validator(
            self.request_schema, resolver=self.schema.resolver)
        self.response_schema = self.schema.resolve('#/definitions/response')
        self.response_validator = jsonschema.Draft4Validator(
            self.response_schema, resolver=self.schema.resolver)
        self.allowed_exceptions = None

    def tearDown(self):
        self.mock_request_patch.stop()

    def call_handle_http(self, args=None, kwargs=None):
        if args is None:
            args = ()
        if kwargs is None:
            kwargs = {}
        return handle_http(self.schema, self.mock_handler, args, kwargs,
                           self.mock_logic, self.request_schema,
                           self.request_validator, self.response_validator,
                           self.allowed_exceptions)

    def mock_logic_exception(self, exception):
        self.mock_request.method = 'POST'
        self.mock_request.content_type = 'application/json; charset=UTF8'
        self.mock_request.mimetype = 'application/json'
        self.mock_request.json = {'a': 'foo', 'b': 1}
        self.mock_logic.side_effect = exception

    def test_from_file(self):
        schema_filepath = os.path.abspath(os.path.join(
            os.path.dirname(__file__), 'schema', 'annotation.yaml'))
        schema_path = os.path.dirname(schema_filepath)
        schema = FlaskResourceSchema.from_file(schema_filepath)
        self.assertIsInstance(schema, FlaskResourceSchema)
        self.assertEqual(schema._schema_path, schema_path)
        self.assertEqual(schema.handle_http, handle_http)
        # Should be able to resolve a ref to another file
        self.assertEqual(schema.resolve('#/properties/auth'), {
            'description': 'auth token',
            'example': 'eb25f25becca416092752b0f457f1271',
            'type': ['string']
        })

    def test_handle_http_with_json(self):
        self.mock_request.method = 'POST'
        self.mock_request.content_type = 'application/json; charset=UTF8'
        self.mock_request.mimetype = 'application/json'
        self.mock_request.json = {'a': 'foo', 'b': 1}
        result = self.call_handle_http((10,), {'b': 2, 'x': 20, 'y': 30})
        self.assertEqual(result, ({'foo': 1}, 201))
        self.assertEqual(self.mock_logic.call_args_list,
                         [mock.call(10, x=20, y=30, a='foo', b=2)])

    def test_handle_http_unsupported_http_method_with_body(self):
        self.mock_request.method = 'GET'
        self.mock_request.content_type = 'application/json; charset=UTF8'
        self.mock_request.mimetype = 'application/json'
        self.mock_request.json = None
        self.mock_request.values = {'a': 'foo', 'b': '1'}
        result = self.call_handle_http((10,))
        expected = ({'foo': 1}, 200)
        self.assertEqual(expected, result)

    def test_handle_http_with_params(self):
        self.mock_request.method = 'DELETE'
        self.mock_request.content_type = 'application/x-www-form-urlencoded'
        self.mock_request.values = {'a': 'foo', 'b': '1'}
        result = self.call_handle_http((10,), {'b': 2, 'x': 20, 'y': 30})
        self.assertEqual(result, ({'foo': 1}, 204))
        self.assertEqual(self.mock_logic.call_args_list,
                         [mock.call(10, x=20, y=30, a='foo', b=2)])

    @mock.patch('doctor.flask.MAX_RESPONSE_LENGTH', 10)
    @mock.patch('doctor.flask.logging')
    def test_handle_http_catch_and_log_response_validaiton_errors(
            self, logging_mock):
        """
        This test verifies that if the schema attribute
        `raise_response_validation_errors` is False we simply log
        the error as a warning and return the result, instead of
        raising the exception.
        """
        self.schema = FlaskResourceSchema({
            'definitions': {
                'a': {'type': 'string'},
                'b': {'type': 'integer'},
                'c': {'type': 'integer'},
                'response': {
                    'type': 'object',
                    'properties': {
                        'foo': {'type': 'integer'},
                    },
                    'additionalProperties': False
                }
            }
        }, raise_response_validation_errors=False)
        self.mock_logic = mock.Mock(spec=['_autospec'])
        self.mock_logic.return_value = {'dinosaurs': 'are extinct'}
        self.mock_request.method = 'GET'
        self.mock_request.path = '/foo'
        self.mock_request.content_type = 'application/x-www-form-urlencoded'
        self.mock_request.values = {'a': 'foo', 'b': '1'}
        result = self.call_handle_http((10,), {'b': 2, 'x': 20, 'y': 30})

        expected = ({'dinosaurs': 'are extinct'}, 200)
        self.assertEqual(expected, result)
        expected_call = mock.call(
            'Response to %s %s does not validate: %s.', 'GET', '/foo',
            "{'dinosaur...", exc_info=mock.ANY)
        self.assertEqual(expected_call, logging_mock.warning.call_args)

    def test_handle_http_raise_response_validaiton_errors(self):
        """
        This test verifies that if the schema attribute
        `raise_response_validation_errors` is True we raise the
        exception.
        """
        self.schema = FlaskResourceSchema({
            'definitions': {
                'a': {'type': 'string'},
                'b': {'type': 'integer'},
                'c': {'type': 'integer'},
                'response': {
                    'type': 'object',
                    'properties': {
                        'foo': {'type': 'integer'},
                    },
                    'additionalProperties': False
                }
            }
        }, raise_response_validation_errors=True)
        self.mock_logic = mock.Mock(spec=['_autospec'])
        self.mock_logic.return_value = {'dinosaurs': 'are extinct'}
        self.mock_request.method = 'GET'
        self.mock_request.content_type = 'application/x-www-form-urlencoded'
        self.mock_request.values = {'a': 'foo', 'b': '1'}
        with self.assertRaises(HTTP400Exception) as context:
            self.call_handle_http((10,), {'b': 2, 'x': 20, 'y': 30})
        errors = context.exception.errobj
        expected = {
            '_other': ("Additional properties are not allowed ('dinosaurs' was "
                       "unexpected)"),
        }
        self.assertEqual(expected, errors)

    def test_handle_http_raise_response_validaiton_errors_response_class(self):
        """
        This test verifies that if the schema attribute
        `raise_response_validation_errors` is True we raise the
        exception.  This specifically tests when the response is an instance
        of ~doctor.response.Response
        """
        self.schema = FlaskResourceSchema({
            'definitions': {
                'a': {'type': 'string'},
                'b': {'type': 'integer'},
                'c': {'type': 'integer'},
                'response': {
                    'type': 'object',
                    'properties': {
                        'foo': {'type': 'integer'},
                    },
                    'additionalProperties': False
                }
            }
        }, raise_response_validation_errors=True)
        self.mock_logic = mock.Mock(spec=['_autospec'])
        self.mock_logic.return_value = Response({'dinosaurs': 'are extinct'})
        self.mock_request.method = 'GET'
        self.mock_request.content_type = 'application/x-www-form-urlencoded'
        self.mock_request.values = {'a': 'foo', 'b': '1'}
        with self.assertRaises(HTTP400Exception) as context:
            self.call_handle_http((10,), {'b': 2, 'x': 20, 'y': 30})
        errors = context.exception.errobj
        expected = {
            '_other': ("Additional properties are not allowed ('dinosaurs' was "
                       "unexpected)"),
        }
        self.assertEqual(expected, errors)

    def test_handle_http_no_schemas(self):
        result = handle_http(self.schema, self.mock_handler, (1, 2),
                             {'a': 3, 'b': 4}, self.mock_logic, None, None,
                             None, None)
        self.assertEqual(result, ({'foo': 1}, 200))
        self.assertEqual(self.mock_logic.call_args_list,
                         [mock.call(1, 2, a=3, b=4)])

    def test_handle_http_response(self):
        """
        Verifies we return the correct tuple from handle_http if the result
        of the logic function is a `Response` instance.
        """
        result = handle_http(self.schema, self.mock_handler, (1, 2),
                             {'a': 3, 'b': 4}, self.mock_flask_response_logic,
                             None, None, None, None)
        expected = ('1,2\n3,4', 200, self.response_headers)
        self.assertEqual(expected, result)

    def test_http_error(self):
        """Schematic errors should be re-raised as a SchematicHTTPException."""
        self.mock_request.content_type = 'application/x-www-form-urlencoded'
        self.mock_request.values = {'b': 'bad integer'}
        with self.assertRaisesRegexp(
                SchematicHTTPException, r'b must be a valid type \(integer\)'):
            self.call_handle_http()

    def test_handle_invalid_value_error(self):
        self.mock_logic_exception(InvalidValueError('400'))
        with self.assertRaisesRegexp(HTTP400Exception, r'400'):
            self.call_handle_http()

    def test_handle_immutable_error(self):
        self.mock_logic_exception(ImmutableError('409'))
        with self.assertRaisesRegexp(HTTP409Exception, r'409'):
            self.call_handle_http()

    def test_handle_parse_error(self):
        self.mock_logic_exception(ParseError('400'))
        with self.assertRaisesRegexp(HTTP400Exception, r'400'):
            self.call_handle_http()

    def test_handle_schema_validation_error(self):
        self.mock_logic_exception(SchemaValidationError('400'))
        with self.assertRaisesRegexp(HTTP400Exception, r'400'):
            self.call_handle_http()

    def test_handle_unauthorized_error(self):
        self.mock_logic_exception(UnauthorizedError('401'))
        with self.assertRaisesRegexp(HTTP401Exception, r'401'):
            self.call_handle_http()

    def test_handle_forbidden_error(self):
        self.mock_logic_exception(ForbiddenError('403'))
        with self.assertRaisesRegexp(HTTP403Exception, r'403'):
            self.call_handle_http()

    def test_handle_not_found_error(self):
        self.mock_logic_exception(NotFoundError('404'))
        with self.assertRaisesRegexp(HTTP404Exception, r'404'):
            self.call_handle_http()

    @mock.patch('doctor.flask.current_app')
    def test_handle_unexpected_error(self, mock_current_app):
        mock_current_app.config = {'DEBUG': False}
        self.mock_logic_exception(TypeError('bad type'))
        with self.assertRaisesRegexp(
                HTTP500Exception, r'Uncaught error in logic function'):
            self.call_handle_http()

        # When DEBUG is True, it should reraise the original exception
        mock_current_app.config = {'DEBUG': True}
        with self.assertRaisesRegexp(TypeError, 'bad type'):
            self.call_handle_http()

    @mock.patch('doctor.flask.current_app')
    def test_handle_unexpected_error_allowed_exceptions(self, mock_current_app):
        mock_current_app.config = {'DEBUG': False}

        class ExceptionalException(Exception):
            pass
        self.allowed_exceptions = [ExceptionalException]

        # allowed exceptions should be re-raised
        self.mock_logic_exception(ExceptionalException('should be re-raised'))
        with self.assertRaisesRegexp(
                ExceptionalException, r'should be re-raised'):
            self.call_handle_http()

        # but other exceptions should still be caught
        self.mock_logic_exception(TypeError('bad type'))
        with self.assertRaisesRegexp(
                HTTP500Exception, r'Uncaught error in logic function'):
            self.call_handle_http()



import inspect
import os
from functools import wraps

import mock
import six

from .base import TestCase

from doctor.utils import (
    exec_params, get_description_lines, get_module_attr, nested_set,
    undecorate_func)


if six.PY2:
    getargspec_func = inspect.getargspec
else:
    getargspec_func = inspect.getfullargspec


def does_nothing(func):
    """An example decorator that does nothing."""
    @wraps(func)
    def wrapper(*args, **kwargs):
        return func(*args, **kwargs)
    return wrapper


def dec_with_args(foo='foo', bar='bar'):
    """An example decorator that takes args and passes value to func."""
    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            return func('arg', *args, **kwargs)
        return wrapper
    return decorator


def dec_with_args_one_of_which_allows_a_func(foo, bar=None):
    """An example decorator that takes args where one is a function.

    In this case, bar accepts a func which transforms foo.
    """
    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            if bar is not None:
                bar(foo)
            return func(*args, **kwargs)
        return wrapper
    return decorator


class TestUtils(TestCase):

    def test_exec_params_function(self):
        def logic(a, b, c=None):
            return a+1, b+1, c
        logic._argspec = getargspec_func(logic)
        kwargs = {'c': 1, 'd': 'a'}
        args = (1, 2)
        a, b, c = exec_params(logic, *args, **kwargs)
        self.assertEqual(a, args[0]+1)
        self.assertEqual(b, args[1]+1)
        self.assertEqual(c, kwargs['c'])

    def test_exec_params_callable(self):
        kwargs = {'c': 1, 'd': 'a'}
        args = (1, 2)

        class Foo(object):
            def __call__(self, a, b, c=None):
                return a+1, b+1, c
        logic = Foo()
        logic._argspec = getargspec_func(logic.__call__)
        a, b, c = exec_params(logic, *args, **kwargs)
        self.assertEqual(a, args[0]+1)
        self.assertEqual(b, args[1]+1)
        self.assertEqual(c, kwargs['c'])

    def test_exec_params_args_only(self):
        def logic(a, b, c):
            return a+1, b+1, c
        logic._argspec = getargspec_func(logic)
        kwargs = {'d': 1, 'e': 2}
        args = (1, 2, 3)
        a, b, c = exec_params(logic, *args, **kwargs)
        self.assertEqual(a, args[0]+1)
        self.assertEqual(b, args[1]+1)
        self.assertEqual(c, 3)

    def test_exec_params_kwargs_only(self):
        def logic(a=1, b=2, c=3):
            return a+1, b+1, c
        logic._argspec = getargspec_func(logic)
        kwargs = {'d': 1, 'e': 2}
        args = (1, 2, 3)
        a, b, c = exec_params(logic, *args, **kwargs)
        self.assertEqual(a, args[0]+1)
        self.assertEqual(b, args[1]+1)
        self.assertEqual(c, 3)

    def test_exec_params_with_magic_kwargs(self):
        """
        This test replicates a bug where if a logic function were decorated
        and we sendt a json body request that contained parameters not in
        the logic function's signature it would cause a TypeError similar to
        the following for the logic function defined in this test:

        TypeError: logic() got an unexpected keyword argument 'e'

        This test verifies if a logic function uses the **kwargs functionality
        that we pass along those to the logic function in addition to any
        other positional or keyword arguments.
        """
        def logic(a, b=2, c=3, **kwargs):
            return (a, b, c, kwargs)

        logic._argspec = getargspec_func(logic)
        kwargs = {'c': 3, 'e': 2}
        args = (1, 2)
        actual = exec_params(logic, *args, **kwargs)
        self.assertEqual((1, 2, 3, {'e': 2}), actual)

    @mock.patch('doctor.utils.open',
                new_callable=mock.mock_open,
                read_data='mock_attr = "something"')
    @mock.patch('doctor.utils.compile', create=True)
    @mock.patch('doctor.utils.os', autospec=True)
    @mock.patch('doctor.utils.sys', autospec=True)
    def test_get_module_attr(self, mock_sys, mock_os, mock_compile, m):
        def side_effect(source, filename, mode, flags=0, dont_inherit=False):
            self.assertEqual(mock_sys.path, ['foo', 'bar', '/foo/bar'])
            return compile(source, filename, mode, flags=flags,
                           dont_inherit=dont_inherit)
        mock_compile.side_effect = side_effect
        mock_os.getcwd.return_value = mock.sentinel.old_cwd
        mock_os.path = os.path
        mock_sys.path = ['foo', 'bar']
        namespace = {}
        result = get_module_attr('/foo/bar/baz', 'mock_attr',
                                 namespace=namespace)
        m.assert_called_once_with('/foo/bar/baz', 'r')
        mock_compile.assert_called_once_with('mock_attr = "something"',
                                             '/foo/bar/baz', 'exec')
        self.assertEqual(result, 'something')
        self.assertEqual(namespace['__file__'], '/foo/bar/baz')
        self.assertEqual(mock_os.chdir.call_args_list,
                         [mock.call('/foo/bar'),
                          mock.call(mock.sentinel.old_cwd)])
        self.assertEqual(mock_sys.path, ['foo', 'bar'])

    def test_get_description_lines(self):
        """
        Tests that get_description_lines properly dedents docstrings and strips
        out values we don't want.
        """
        docstring = """line one

                       line two

                           indented line three

                       line four

                       :param str c: example param
                       :returns: d
                       """
        self.assertEqual(get_description_lines(docstring), [
            'line one',
            '',
            'line two',
            '',
            '    indented line three',
            '',
            'line four',
            '',
        ])

    def test_get_description_lines_none(self):
        """It should just return an empty list for None."""
        self.assertEqual(get_description_lines(None), [])

    def test_get_description_lines_trailing_newline(self):
        """It should add a trailing line if necessary."""
        self.assertEqual(get_description_lines('foo\n:arg'), ['foo', ''])

    def test_nested_set(self):
        data_dict = {}
        nested_set(data_dict, ["b", "v", "new"], {})
        self.assertEqual(data_dict['b']['v']['new'], {})

        nested_set(data_dict, ["b", "v", "new"], '4')
        self.assertEqual(data_dict['b']['v']['new'], '4')

        nested_set(data_dict, ["a", "v", "new"], '1')
        self.assertEqual(data_dict['a']['v']['new'], '1')

        nested_set(data_dict, ["a", "new"], 'new')
        self.assertEqual(data_dict['a']['new'], 'new')

    def test_undecorate_func(self):
        def foobar(a, b=False):
            pass

        # No decorator just returns the function
        actual = undecorate_func(foobar)
        self.assertEqual(foobar, actual)

        # Normal decorator with no args
        decorated = does_nothing(foobar)
        actual = undecorate_func(decorated)
        self.assertEqual(foobar, actual)

        # Ensure it can handle multiple decorators
        double_decorated = does_nothing(does_nothing(foobar))
        actual = undecorate_func(double_decorated)
        self.assertEqual(foobar, actual)

        # Ensure it works with decorators that take arguments
        decorated_with_args = dec_with_args('foo1')(foobar)
        actual = undecorate_func(decorated_with_args)
        self.assertEqual(foobar, actual)

        def bar(foo):
            return 'foo ' + foo

        decorated_with_arg_takes_func = (
            dec_with_args_one_of_which_allows_a_func('foo', bar)(foobar))
        actual = undecorate_func(decorated_with_arg_takes_func)
        self.assertEqual(foobar, actual)



import json

import mock
from sphinx.errors import SphinxError
from werkzeug.routing import Rule

from doctor.docs import base
from doctor.errors import SchemaError
from doctor.resource import ResourceSchema, ResourceSchemaAnnotation
from .base import TestCase


class TestDocsBase(TestCase):

    def setUp(self):
        super(TestDocsBase, self).setUp()
        self.schema = ResourceSchema({
            'definitions': {
                'a': {
                    'type': 'string',
                    'description': 'example description for a',
                    'example': 'example string',
                },
                'b': {
                    'type': 'integer',
                    'description': 'example description for b',
                    'example': 123,
                },
                'c': {
                    'type': 'boolean',
                    'description': 'example description for c',
                    'example': True,
                },
                'd': {
                    'type': ['integer', 'null'],
                    'description': 'example description for d',
                    'example': 456,
                },
                'e': {
                    'type': ['array'],
                    'description': 'example description for e',
                    'example': ['example', 'array'],
                },
                'f': {
                    'type': 'object',
                    'description': 'example description for f',
                    'properties': {
                        'str': {
                            'type': 'string',
                            'description': 'A string',
                            'example': 'A String',
                        },
                    },
                    'example': {'str': 'example string'},
                },
                'g': {
                    'type': 'string',
                    'description': 'example description for g',
                    'example': {
                        'oneOf': [
                            'example string 1',
                            'example string 2',
                        ],
                    },
                },
                'h': {
                    'type': ['null'],
                    'description': 'example description for h',
                    'example': None
                },
                'i': {
                    'type': 'array',
                    'description': 'example description for i',
                    'items': {
                        '$ref': '#/definitions/b',
                    },
                },
                # An example with an array of objects with no example
                # key on the object.
                'j': {
                    'type': 'array',
                    'description': 'An array of objects',
                    'items': {
                        'type': 'object',
                        'properties': {
                            'str': {
                                'type': 'string',
                                'description': 'A String',
                                'example': 'string',
                            },
                            'int': {
                                'type': 'integer',
                                'description': 'An integer',
                                'example': 1022,
                            },
                        },
                        'additionalProperties': False,
                    },
                },
                'k': {
                    'type': 'string',
                    'description': 'The type of shirt.',
                    'enum': ['tshirt', 'sweater'],
                    'example': 'tshirt',
                },
                # An object w/o any properties.
                'l': {
                    'type': 'object',
                    'description': 'An object that can have any properties',
                    'example': {
                        'foo': 'bar',
                    },
                },
                # An object with a property that is a reference we can't resolve
                'm': {
                    'type': 'object',
                    'description': 'An object with a propertery that is a ref',
                    'example': {
                        'hey': 'ref',
                    },
                    'properties': {
                        'hey': {
                            '$ref': 'dinosaurs.yaml#/definitions/trex',
                        },
                    },
                },
                # An object with a property to document whose type is an
                # array, rather than a single item string.
                'n': {
                    'type': ['null', 'object'],
                    'description': 'example description for n',
                    'properties': {
                        'str': {
                            'type': 'string',
                            'description': 'A string',
                            'example': 'A String',
                        },
                    },
                    'example': {'str': 'example string `n`'},
                },
            },
        })
        self.request_schema = self.schema._create_request_schema(
            params=('a', 'b', 'c'), required=('a', 'b'))
        self.response_schema = {
            'type': 'array',
            'items': {
                'type': 'object',
                'properties': {
                    'b': {'$ref': '#/definitions/b'},
                    'c': {'$ref': '#/definitions/c'},
                },
            },
        }

        def mock_logic():
            pass
        mock_logic.required_args = ['a', 'c']

        self.annotation = ResourceSchemaAnnotation(
            logic=mock_logic, http_method='GET', schema=self.schema,
            request_schema=self.request_schema,
            response_schema=self.response_schema)

    def test_get_example_value(self):
        definitions = self.schema.resolve('#/definitions')
        results = [base.get_example_value(prop_schema, self.schema, key)
                   for key, prop_schema in sorted(definitions.items())]
        expected = [
            'example string',
            123,
            True,
            456,
            ['example', 'array'],
            {'str': 'example string'},
            'example string 1',
            None,
            [123],
            [{'str': 'string', 'int': 1022}],
            'tshirt',
            {'foo': 'bar'},
            {'hey': 'ref'},
            {'str': 'example string `n`'},
        ]
        self.assertEqual(expected, results)

    def test_get_example_for_object(self):
        """
        This test verifies we can get an example for an object that is
        missing an example from normal properties.
        """
        schema = {
            'definitions': {
                'a': {
                    'type': 'object',
                    'properties': {
                        'foo': {
                            'type': 'string',
                            'description': 'A foo',
                            'example': 'Foo!',
                        },
                    },
                    'additionalProperties': False,
                },
            },
        }
        schema = ResourceSchema(schema)
        definitions = schema.resolve('#/definitions')
        result = base.get_example_for_object(definitions['a'], schema)
        expected = {'foo': 'Foo!'}
        self.assertEqual(expected, result)

    def test_get_example_for_object_with_different_schema(self):
        """
        This test verifies if the object_schema is a different file than the
        schema argument, that we create a new ResourceSchema instance from
        the object schema in order to resolve example values.
        """
        schema = ResourceSchema({
            'definitions': {
                'a': {
                    'type': 'object',
                    'properties': {
                        'foo': {
                            'type': 'string',
                            'description': 'A foo',
                            'example': 'Foo!',
                        },
                    },
                    'additionalProperties': False,
                },
            },
        })
        object_schema = {
            '$schema': 'http://json-schema.org/draft-04/schema#',
            'title': 'Another type of schema',
            'description': 'A description',
            'definitions': {
                'interesting_thing': {
                    'type': 'object',
                    'properties': {
                        'cat': {
                            'type': 'string',
                            'description': 'The name of a cat',
                            'example': 'Els',
                        },
                    },
                },
                'interesting_things': {
                    'type': 'array',
                    'description': 'an array of intersting things',
                    'items': {
                        '$ref': '#/definitions/interesting_thing',
                    },
                }
            },
            'type': 'object',
            'properties': {
                'interesting_things': {
                    '$ref': '#/definitions/interesting_things',
                },
            }
        }
        actual = base.get_example_for_object(object_schema, schema)
        expected = {'interesting_things': [{'cat': 'Els'}]}
        self.assertEqual(expected, actual)

    def test_get_example_value_anyof(self):
        """
        This test verifies that if a property has an anyOf declaration for the
        example, that we first resolve the anyOf by grabbing the first
        defintion in it's list.
        """
        property_schema = {
            'anyOf': [
                {'$ref': '#/definitions/b'},
                {'$ref': '#/definitions/c'}
            ]
        }
        actual = base.get_example_value(property_schema, self.schema, 'type')
        # expected value should be the example from definitions/b
        expected = 123
        self.assertEqual(expected, actual)

    def test_get_example_reraises_with_key(self):
        """
        This test verifies we re-raise the SchemaError with the offending key.
        """
        property_schema = {
            'type': 'integer',
            'description': 'An integer and we are missing the example'
        }
        expected_message = (
            r"Error resolving #example from \{'type': 'integer', "
            r"'description': 'An integer and we are missing the "
            r"example'\} for property `foobar`\. Unresolvable JSON "
            r"pointer: u?'example' \(from \)"
        )
        with self.assertRaisesRegexp(SchemaError, expected_message):
            base.get_example_value(property_schema, self.schema, 'foobar')

    def test_get_example_lines_json(self):
        """Tests an example when the response is valid JSON."""
        headers = {'GeoIp-Country-Code': 'US'}
        lines = base.get_example_lines(headers, 'GET', 'http://example.com/',
                                       {}, json.dumps({'foo': 1, 'bar': 2}))
        self.assertEqual(lines, [
            '',
            'Example Request:',
            '',
            '.. code-block:: bash',
            '',
            '   curl http://example.com/ -X GET -H \'GeoIp-Country-Code: US\'',
            '',
            'Example Response:',
            '',
            '.. code-block:: json',
            '',
            '   {',
            '     "bar": 2,',
            '     "foo": 1',
            '   }',
        ])

    def test_get_example_lines_text(self):
        """Tests an example when the response is *not* valid JSON."""
        lines = base.get_example_lines({}, 'GET', 'http://example.com/', {},
                                       'hello, world!')
        self.assertEqual(lines, [
            '',
            'Example Request:',
            '',
            '.. code-block:: bash',
            '',
            '   curl http://example.com/ -X GET',
            '',
            'Example Response:',
            '',
            '.. code-block:: text',
            '',
            '   hello, world!',
        ])

    def test_get_json_object_lines(self):
        result = base.get_json_object_lines(
            self.annotation, self.request_schema, field='>json', url_params=[])
        self.assertEqual(result, [
            ':>json str a: example description for a',
            ':>json int b: example description for b',
            ':>json bool c: example description for c',
        ])

    def test_get_json_object_lines_for_request_with_enum(self):
        request_schema = self.schema._create_request_schema(
            params=('a', 'b', 'c', 'k'), required=('a', 'b'))
        result = base.get_json_object_lines(
            self.annotation, request_schema, field='>json', url_params=[],
            request=True)
        self.assertEqual(result, [
            ':>json str a: **Required**.  example description for a',
            ':>json bool c: **Required**.  example description for c',
            ':>json int b: example description for b',
            (":>json str k: The type of shirt. Must be one of: "
             "`['tshirt', 'sweater']`."),
        ])

    def test_get_json_object_lines_for_request(self):
        """
        This tests that when the request kwarg is True that any
        required params have the description prefixed with
        **Required** and sorted in alphabetical order, followed by
        any optional parameters in alpabetical order.
        """
        url_params = ['a']
        result = base.get_json_object_lines(
            self.annotation, self.request_schema, field='>json',
            url_params=url_params, request=True)
        self.assertEqual(result, [
            ':param str a: **Required**.  example description for a',
            ':>json bool c: **Required**.  example description for c',
            ':>json int b: example description for b',
        ])

    def test_get_json_object_lines_object(self):
        def mock_logic():
            pass
        mock_logic.required_args = ['f']
        request_schema = self.schema._create_request_schema(
            params=('f',), required=('f',))
        annotation = ResourceSchemaAnnotation(
            logic=mock_logic, http_method='GET', schema=self.schema,
            request_schema=self.request_schema,
            response_schema=self.response_schema)
        result = base.get_json_schema_lines(
            annotation, request_schema, field='>json', route='/foo')
        expected = [
            ':>json dict f|object: example description for f',
            ':>json str str|objectproperty: A string'
        ]
        self.assertEqual(expected, result)

    def test_get_json_object_lines_object_type_is_array(self):
        """
        This test verifies if a definition has a type of array and one of the
        types is an object that we still docuemnt the properties of the object.
        This is a regression test where we used to only document object
        properties if the type was a string equal to 'object'.
        """
        def mock_logic():
            pass
        mock_logic.required_args = ['n']
        request_schema = self.schema._create_request_schema(
            params=('n',), required=('n',))
        annotation = ResourceSchemaAnnotation(
            logic=mock_logic, http_method='GET', schema=self.schema,
            request_schema=self.request_schema,
            response_schema=self.response_schema)
        result = base.get_json_schema_lines(
            annotation, request_schema, field='>json', route='/foo')
        expected = [
            ':>json null,dict n|object: example description for n',
            ':>json str str|objectproperty: A string'
        ]
        self.assertEqual(expected, result)

    def test_get_json_object_lines_object_no_properties(self):
        """
        This is a regression test that ensures we don't blow up if we encounter
        an object that doesn't have any defined properties to document.
        """
        def mock_logic():
            pass
        mock_logic.required_args = ['l']
        request_schema = self.schema._create_request_schema(
            params=('l',), required=('l',))
        annotation = ResourceSchemaAnnotation(
            logic=mock_logic, http_method='GET', schema=self.schema,
            request_schema=self.request_schema,
            response_schema=self.response_schema)
        result = base.get_json_schema_lines(
            annotation, request_schema, field='>json', route='/foo')
        expected = [':>json dict l: An object that can have any properties']
        self.assertEqual(expected, result)

    def test_get_json_object_lines_object_property_is_a_ref(self):
        """
        This is a regression test that ensures we don't blow up if we encounter
        an object that has a property that can't be resolved.
        """
        def mock_logic():
            pass
        mock_logic.required_args = ['m']
        request_schema = self.schema._create_request_schema(
            params=('m',), required=('m',))
        annotation = ResourceSchemaAnnotation(
            logic=mock_logic, http_method='GET', schema=self.schema,
            request_schema=self.request_schema,
            response_schema=self.response_schema)
        result = base.get_json_schema_lines(
            annotation, request_schema, field='>json', route='/foo')
        expected = [':>json dict m: An object with a propertery that is a ref']
        self.assertEqual(expected, result)

    def test_get_json_schema_lines(self):
        result = base.get_json_schema_lines(
            self.annotation, self.request_schema, field='>json', route='/foo')
        self.assertEqual(result, [
            ':>json str a: example description for a',
            ':>json int b: example description for b',
            ':>json bool c: example description for c',
        ])

    def test_get_json_schema_lines_array(self):
        result = base.get_json_schema_lines(
            self.annotation, self.response_schema, field='<json', route='/foo')
        self.assertEqual(result, [
            ':<jsonarr int b: example description for b',
            ':<jsonarr bool c: example description for c',
        ])

    @mock.patch('doctor.docs.base.logging.warn')
    def test_get_json_schema_lines_array_invalid_type(self, mock_warn):
        self.response_schema['items']['type'] = 'string'
        result = base.get_json_schema_lines(
            self.annotation, self.response_schema, field='<json', route='/foo')
        self.assertEqual(result, [])
        self.assertEqual(mock_warn.call_args_list, [
            mock.call('Not documenting list item of type %s (for %s %s)',
                      'string', 'GET', '/foo')])

    @mock.patch('doctor.docs.base.logging.warn')
    def test_get_json_schema_lines_invalid_type(self, mock_warn):
        result = base.get_json_schema_lines(
            self.annotation, {'type': 'string'}, field='<json', route='/foo')
        self.assertEqual(result, [])
        self.assertEqual(mock_warn.call_args_list, [
            mock.call('Not documenting schema of type %s (for %s %s)',
                      'string', 'GET', '/foo')])

    def test_get_name(self):
        mock_class = mock.Mock()
        mock_class.__module__ = 'foo.bar'
        mock_class.__name__ = 'baz'
        self.assertEqual(base.get_name(mock_class), 'foo.bar.baz')
        mock_class.__module__ = '__builtin__'
        self.assertEqual(base.get_name(mock_class), 'baz')

    def test_resolve(self):
        definitions = self.schema.resolve('#/definitions')
        self.assertEqual(base.resolve('#a', definitions, self.annotation),
                         definitions['a'])
        with self.assertRaises(SphinxError):
            base.resolve('#invalid_ref', definitions, self.annotation)

    def test_resolve_with_oneof(self):
        """
        This test verifies that if a oneOf definition hits the resolve
        method, that we first resolve the oneOf by grabbing the first
        defintion in it's list.
        """
        fragment = 'type'
        definitions = {
            'oneOf': [
                {'$ref': '#/definitions/a'},
                {'$ref': '#/definitions/g'}
            ]
        }
        actual = base.resolve(fragment, definitions, self.annotation)

        expected = {
            'type': 'string',
            'description': 'example description for a',
            'example': 'example string',
        }
        self.assertEqual(expected, actual)

    def test_resolve_with_anyof(self):
        """
        This test verifies that if a anyOf definition hits the resolve
        method, that we first resolve the anyOf by grabbing the first
        defintion in it's list.
        """
        fragment = 'type'
        definitions = {
            'anyOf': [
                {'$ref': '#/definitions/b'},
                {'$ref': '#/definitions/c'}
            ]
        }
        actual = base.resolve(fragment, definitions, self.annotation)

        expected = {
            'type': 'integer',
            'description': 'example description for b',
            'example': 123,
        }
        self.assertEqual(expected, actual)

    def test_resolve_of_type(self):
        schema = {
            'oneOf': [
                {'$ref': '#/definitions/a'},
                {'$ref': '#/definitions/g'}
            ]
        }
        actual = base.resolve_of_type(self.annotation, schema)
        expected = {
            'type': 'string',
            'description': 'example description for a',
            'example': 'example string',
        }
        self.assertEqual(expected, actual)


class TestDocsBaseHarness(TestCase):

    def test_init(self):
        harness = base.BaseHarness('http://foo/')
        self.assertEqual(harness.url_prefix, 'http://foo')

    def test_get_annotation_heading_schematic_title(self):
        """
        This test verifies we use the schematic_title attribute of the
        handler if it is present.
        """
        handler = mock.Mock(schematic_title='Test Title')
        route = '^foo/?$'
        harness = base.BaseHarness('http://foo/')
        actual = harness._get_annotation_heading(handler, route)
        self.assertEqual('Test Title', actual)

    def test_get_annotation_heading_class_path(self):
        """
        This test verifies if the class path has a resource name in it,
        that we use it for the heading.
        e.g. class <api.handlers.foo_bar.FooListHandler> becomes `Foo Bar`
        """
        handler = mock.Mock(spec_set=base.BaseHarness)
        handler.__str__ = mock.Mock(
            return_value='<class api.foo_bar.FooBarListHandler>')
        route = '^foo_bar/?$'
        harness = base.BaseHarness('http://foo/')
        actual = harness._get_annotation_heading(handler, route)

        expected = 'Foo Bar'
        self.assertEqual(expected, actual)

    def test_get_annotation_heading_generic_handlers(self):
        """
        This test verifies if our handlers are not in their own resource
        modules that we get the heading from the handler class name.
        e.g. class <api.handlers.handlers.FooBarListHandler> becomes `Foo Bar`
        """
        handler = mock.Mock(spec_set=base.BaseHarness)
        handler.__str__ = mock.Mock(
            return_value='<class api.handlers.FooBarListHandler>')
        route = '^foo_bar/?$'

        harness = base.BaseHarness('http://foo/')
        actual = harness._get_annotation_heading(handler, route)

        expected = 'Foo Bar'
        self.assertEqual(expected, actual)

    def test_get_annotation_heading_class_name_only(self):
        """
        This test verifies that if our handler has no path and is just
        the class name that we get the heading from the name.
        e.g. class <FooBarListHandler> becomes `Foo Bar`
        """
        handler = mock.Mock(spec_set=base.BaseHarness)
        handler.__str__ = mock.Mock(
            return_value='<class FooBarListHandler>')
        route = '^foo_bar/?$'

        harness = base.BaseHarness('http://foo/')
        actual = harness._get_annotation_heading(handler, route)

        expected = 'Foo Bar'
        self.assertEqual(expected, actual)

    def test_get_annotation_heading_class_path_internal(self):
        """
        This test verifies the path where the class path has the resource
        name in it and it's an internal route.
        """
        handler = mock.Mock(spec_set=base.BaseHarness)
        handler.__str__ = mock.Mock(
            return_value='<class api.foo_bar.InternalFooBarListHandler>')
        route = '^internal/r/foo_bar/?$'
        harness = base.BaseHarness('http://foo/')
        actual = harness._get_annotation_heading(handler, route)

        expected = 'Foo Bar (Internal)'
        self.assertEqual(expected, actual)

    def test_get_annotation_heading_class_name_only_internal(self):
        """
        This test verifies the path where our handler has no path and just
        the class name and that class is internal.
        """
        handler = mock.Mock(spec_set=base.BaseHarness)
        handler.__str__ = mock.Mock(
            return_value='<class InternalFooBarListHandler>')
        route = '^/internal/foo_bar/?$'

        harness = base.BaseHarness('http://foo/')
        actual = harness._get_annotation_heading(handler, route)

        expected = 'Foo Bar (Internal)'
        self.assertEqual(expected, actual)

    def test_get_example_values_get_http_method_with_list_and_dict_vals(self):
        """
        This test verifies if the route we are generating example values for
        is a GET endpoint and the example values are lists or dicts, that we
        json.dumps them when they are returned.  If the HTTP method is any
        other method, they will be returned as normal lists or dicts.
        """
        def mock_logic():
            pass

        route = Rule('/foo/bar/')
        schema = ResourceSchema({
            'definitions': {
                'e': {
                    'type': ['array'],
                    'description': 'example description for e',
                    'example': ['example', 'array'],
                },
                'f': {
                    'type': 'object',
                    'description': 'example description for f',
                    'properties': {
                        'str': {
                            'type': 'string',
                            'description': 'A string',
                            'example': 'A String',
                        },
                    },
                    'example': {'str': 'example string'},
                },
            }
        })
        request_schema = schema._create_request_schema(
            params=('e', 'f'), required=('e', 'f'))

        annotation = ResourceSchemaAnnotation(
            logic=mock_logic, http_method='GET', schema=schema,
            request_schema=request_schema, response_schema=None)
        harness = base.BaseHarness('http://foo/bar/')
        example_values = harness._get_example_values(route, annotation)
        expected = {
            'e': json.dumps(['example', 'array']),
            'f': json.dumps({'str': 'example string'}),
        }
        self.assertEqual(expected, example_values)

        # Change the http method to something other than GET, they should
        # not be json dumped.
        annotation.http_method = 'POST'
        example_values = harness._get_example_values(route, annotation)
        expected = {
            'e': ['example', 'array'],
            'f': {'str': 'example string'},
        }
        self.assertEqual(expected, example_values)



from functools import wraps

import mock
import six

from doctor.resource import ResourceSchema, ResourceSchemaAnnotation
from doctor.flask import handle_http
from .base import TestCase


def does_nothing(func):
    """An example decorator that does nothing."""
    @wraps(func)
    def wrapper(*args, **kwargs):
        return func(*args, **kwargs)
    return wrapper


def plain_logic(a, b=1):
    return (a, b)


def plain_logic_with_kwargs(a, b=1, **kwargs):
    return (a, b, kwargs)


@does_nothing
def decorated_logic(a, b=1):
    return (a, b)


@does_nothing
def decorated_logic_with_kwargs(a, b=1, **kwargs):
    return (a, b, kwargs)


class TestResourceSchema(TestCase):

    def setUp(self):
        self.mock_handle_http = mock.Mock(autospec=True, spec=handle_http,
                                          return_value=mock.sentinel.result)

    def test_create_request_schema(self):
        schema = ResourceSchema({'definitions': mock.sentinel.definitions},
                                self.mock_handle_http)
        request_schema = schema._create_request_schema(params=('a', 'b', 'c'),
                                                       required=('b', 'c'))
        self.assertEqual(request_schema, {
            'additionalProperties': True,
            'definitions': mock.sentinel.definitions,
            'properties': {'a': {'$ref': '#/definitions/a'},
                           'b': {'$ref': '#/definitions/b'},
                           'c': {'$ref': '#/definitions/c'}},
            'required': ('b', 'c'),
            'type': 'object',
        })

    def test_parse_params(self):
        schema = ResourceSchema({
            'definitions': {
                'a': {'type': 'string'},
                'b': {'type': 'integer'},
                'c': {'type': 'integer'},
                'd': {'type': ['integer', 'null']},
                'e': {'type': ['array']},
                'f': {'type': ['object']}
            }
        }, self.mock_handle_http)
        request_schema = schema._create_request_schema(
            params=('a', 'b', 'c', 'd', 'e', 'f'),
            required=None)
        params = {
            'a': 'hodor',
            'b': '1',
            'd': '',
            'e': '[1, 2]',
            'f': '{"foo": 1}',
        }
        expected = {
            'a': 'hodor',
            'b': 1,
            'd': None,
            'e': [1, 2],
            'f': {'foo': 1},
        }
        self.assertEqual(schema._parse_params(params, request_schema),
                         expected)

    def test_create_http_method(self):
        def logic(a, b, c=1):
            """Foo."""
            return a

        s = mock.sentinel
        schema = ResourceSchema({
            'definitions': {
                'a': {'type': 'string'},
                'b': {'type': 'integer'},
                'c': {'type': 'integer'},
                'request': {
                    'type': 'object',
                    'properties': {
                        'bar': {'type': 'integer'},
                    },
                    'additionalProperties': False,
                },
                'response': {
                    'type': 'object',
                    'properties': {
                        'foo': {'type': 'integer'},
                    },
                    'additionalProperties': False,
                },
            },
        }, self.mock_handle_http)
        handler = schema._create_http_method(
            logic, 'POST', params=('a', 'b', 'c'), required=('a', 'b'),
            response='response', allowed_exceptions=s.allowed_exceptions)
        annotation = handler._schema_annotation
        self.assertTrue(callable(handler))
        self.assertEqual(handler.__name__, 'logic')
        self.assertEqual(handler.__doc__, 'Foo.')
        self.assertEqual(annotation.request_schema, {
            'additionalProperties': True,
            'definitions': schema.resolve('#/definitions'),
            'properties': {
                'a': {'$ref': '#/definitions/a'},
                'b': {'$ref': '#/definitions/b'},
                'c': {'$ref': '#/definitions/c'}
            },
            'required': ('a', 'b'),
            'type': 'object',
        })
        self.assertEqual(annotation.response_schema,
                         schema.resolve('#/definitions/response'))
        result = handler(s.handler, 1, x=2, y=3)
        self.assertEqual(result, s.result)
        self.assertEqual(self.mock_handle_http.call_args_list, [
            mock.call(schema, s.handler, (1,), {'x': 2, 'y': 3}, logic,
                      annotation.request_schema, mock.ANY, mock.ANY,
                      s.allowed_exceptions)
        ])

        # Test a handler with an explicit request schema
        self.mock_handle_http.reset_mock()
        handler = schema._create_http_method(logic, 'POST',
                                             request='request',
                                             response='response')
        annotation = handler._schema_annotation
        self.assertEqual(annotation.request_schema,
                         schema.resolve('#/definitions/request'))
        self.assertEqual(annotation.response_schema,
                         schema.resolve('#/definitions/response'))

        # Test a handler without any params
        self.mock_handle_http.reset_mock()
        handler = schema._create_http_method(logic, 'POST',
                                             response='response')
        annotation = handler._schema_annotation
        self.assertIsNone(annotation.request_schema)
        self.assertEqual(annotation.response_schema,
                         schema.resolve('#/definitions/response'))
        result = handler(s.handler, 1, x=2, y=3)
        self.assertEqual(result, s.result)
        self.assertEqual(self.mock_handle_http.call_args_list, [
            mock.call(schema, s.handler, (1,), {'x': 2, 'y': 3}, logic,
                      None, None, mock.ANY, None)
        ])

        # Test a handler without a response schema
        self.mock_handle_http.reset_mock()
        handler = schema._create_http_method(
            logic, 'POST', params=('a', 'b', 'c'), required=('a', 'b'))
        annotation = handler._schema_annotation
        self.assertIsNone(annotation.response_schema)
        handler(s.handler, 1, x=2, y=3)
        self.assertEqual(result, s.result)
        self.assertEqual(self.mock_handle_http.call_args_list, [
            mock.call(schema, s.handler, (1,), {'x': 2, 'y': 3}, logic,
                      annotation.request_schema, mock.ANY, None, None)
        ])

    def test_create_http_method_logic_argspec(self):
        """
        This tests that the the return value's _argspec attribute is set
        correctly when using a normal function and decorated ones.

        This test also replicates a bug where if a logic function were decorated
        and we sent a json body request that contained parameters not in
        the logic function's signature it would cause a TypeError similar to
        the following for the logic function defined in this test:

        TypeError: logic() got an unexpected keyword argument 'e'

        """
        s = mock.sentinel
        schema = ResourceSchema({
            'definitions': {
                'a': {'type': 'string'},
                'b': {'type': 'integer'},
                'c': {'type': 'integer'},
                'request': {
                    'type': 'object',
                    'properties': {
                        'bar': {'type': 'integer'},
                    },
                    'additionalProperties': False,
                },
                'response': {
                    'type': 'object',
                    'properties': {
                        'foo': {'type': 'integer'},
                    },
                    'additionalProperties': False,
                },
            },
        }, self.mock_handle_http)

        handler = schema._create_http_method(
            plain_logic, 'POST', params=('a', 'b'), required=('a',),
            response='response', allowed_exceptions=s.allowed_exceptions)
        spec = handler._argspec
        self.assertEqual(['a', 'b'], spec.args)
        self.assertIsNone(spec.varargs)
        self.assertIsNone(spec.keywords if six.PY2 else spec.varkw)
        self.assertEqual((1,), spec.defaults)

        handler = schema._create_http_method(
            plain_logic_with_kwargs, 'POST', params=('a', 'b'), required=('a',),
            response='response', allowed_exceptions=s.allowed_exceptions)
        spec = handler._argspec
        self.assertEqual(['a', 'b'], spec.args)
        self.assertIsNone(spec.varargs)
        self.assertEqual('kwargs', spec.keywords if six.PY2 else spec.varkw)
        self.assertEqual((1,), spec.defaults)

        # This caused issue #12 on github.
        handler = schema._create_http_method(
            decorated_logic, 'POST', params=('a', 'b'), required=('a',),
            response='response', allowed_exceptions=s.allowed_exceptions)
        spec = handler._argspec
        self.assertEqual(['a', 'b'], spec.args)
        self.assertIsNone(spec.varargs)
        self.assertIsNone(spec.keywords if six.PY2 else spec.varkw)
        self.assertEqual((1,), spec.defaults)

        handler = schema._create_http_method(
            decorated_logic_with_kwargs, 'POST', params=('a', 'b'),
            required=('a',), response='response',
            allowed_exceptions=s.allowed_exceptions)
        spec = handler._argspec
        self.assertEqual(['a', 'b'], spec.args)
        self.assertIsNone(spec.varargs)
        self.assertEqual('kwargs', spec.keywords if six.PY2 else spec.varkw)
        self.assertEqual((1,), spec.defaults)

    def test_before_after_callables(self):
        mock_logic = mock.Mock()
        mock_logic.__name__ = 'mock_name'
        mock_logic.__doc__ = mock.sentinel.__doc__
        mock_logic.__closure__ = None

        schema = ResourceSchema({
            'definitions': {
                'a': {'type': 'string'},
                'request': {
                    'type': 'object',
                    'properties': {
                        'bar': {'type': 'integer'},
                    },
                    'additionalProperties': False,
                },
                'response': {
                    'type': 'object',
                    'properties': {
                        'foo': {'type': 'integer'},
                    },
                    'additionalProperties': False,
                },
            },
        }, self.mock_handle_http)

        before_mock = mock.Mock()
        after_mock = mock.Mock()
        handler = schema.http_post(
            mock_logic,
            before=before_mock,
            after=after_mock)
        handler(mock.sentinel.handler, 1, x=2)
        self.assertTrue(before_mock.called)
        self.assertTrue(after_mock.called)

        before_mock.reset_mock()
        after_mock.reset_mock()
        handler = schema.http_get(
            mock_logic, after=after_mock)
        handler(mock.sentinel.handler, 1, x=2)
        self.assertFalse(before_mock.called)
        self.assertTrue(after_mock.called)

        def after(self, result, *args, **kwargs):
            result.name = 'hi'

        handler = schema.http_get(
            mock_logic, after=after)
        result = handler(mock.sentinel.handler, 1, x=2)
        self.assertEqual(result.name, 'hi')

    @mock.patch.object(ResourceSchema, '_create_http_method', autospec=True,
                       return_value=mock.sentinel.result)
    def test_http_methods(self, mock_create_http_method):
        s = mock.sentinel
        schema = ResourceSchema({}, self.mock_handle_http)
        for method in ('delete', 'get', 'post', 'put'):
            mock_create_http_method.reset_mock()
            fn = getattr(schema, 'http_{}'.format(method))
            result = fn(s.logic, s.request, s.response, s.params, s.required,
                        title=s.title)
            self.assertEqual(result, s.result)
            self.assertEqual(
                mock_create_http_method.call_args_list,
                [mock.call(schema, s.logic, method.upper(), request=s.request,
                           response=s.response, params=s.params,
                           required=s.required, title=s.title, before=None,
                           after=None, allowed_exceptions=None,
                           omit_args=None)])


class TestResourceSchemaAnnotation(TestCase):

    def test_init(self):
        s = mock.sentinel
        annotation = ResourceSchemaAnnotation(
            s.logic, 'POST', s.schema, s.request_schema,
            s.response_schema)
        self.assertEqual(annotation.logic, s.logic)
        self.assertEqual(annotation.http_method, 'POST')
        self.assertEqual(annotation.schema, s.schema)
        self.assertEqual(annotation.request_schema, s.request_schema)
        self.assertEqual(annotation.response_schema, s.response_schema)
        self.assertEqual(annotation.title, 'Create')

    def test_init_title(self):
        tests = (
            # (http_method, title, expected)
            ('GET', None, 'Retrieve'),
            ('POST', None, 'Create'),
            ('PUT', None, 'Update'),
            ('DELETE', None, 'Delete'),
            ('PUT', 'Batch', 'Batch'),
        )
        s = mock.sentinel
        for http_method, title, expected in tests:
            annotation = ResourceSchemaAnnotation(
                s.logic, http_method, s.schema, s.request_schema,
                s.response_schema, title=title)
            self.assertEqual(expected, annotation.title)

    def test_get_annotation(self):
        """Tests that get_annotation works for badly decorated functions."""
        def decorator(fn):
            def wrapper():
                fn()
            return wrapper
        mock_logic = mock.Mock()
        mock_logic._schema_annotation = mock.sentinel.logic_annotation
        wrapper = decorator(mock_logic)
        self.assertEqual(ResourceSchemaAnnotation.get_annotation(mock_logic),
                         mock.sentinel.logic_annotation)
        self.assertEqual(ResourceSchemaAnnotation.get_annotation(wrapper),
                         mock.sentinel.logic_annotation)
        wrapper._schema_annotation = mock.sentinel.wrapper_annotation
        self.assertEqual(ResourceSchemaAnnotation.get_annotation(wrapper),
                         mock.sentinel.wrapper_annotation)
        delattr(wrapper, '_schema_annotation')
        delattr(mock_logic, '_schema_annotation')
        self.assertIsNone(ResourceSchemaAnnotation.get_annotation(wrapper))
